# -*- coding: utf-8; -*-
"""grow category.code

Revision ID: 391fac830cb1
Revises: be079e974a52
Create Date: 2018-08-24 12:49:29.826310

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '391fac830cb1'
down_revision = u'be079e974a52'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # category
    op.alter_column('category', 'code', type_=sa.String(length=20))


def downgrade():

    # category
    op.alter_column('category', 'code', type_=sa.String(length=8))



# -*- coding: utf-8; -*-

from __future__ import unicode_literals, absolute_import

import unittest

from mock import patch, Mock

from rattail import monitoring
from rattail.config import RattailConfig


class TestMonitorAction(unittest.TestCase):

    def setUp(self):
        self.config = RattailConfig()

    def test_attributes(self):
        action = monitoring.MonitorAction(self.config)
        self.assertIs(action.config, self.config)
        self.assertTrue(hasattr(action, 'app'))

    def test_not_implemented(self):
        action = monitoring.MonitorAction(self.config)
        self.assertRaises(NotImplementedError, action)


class TestCommandAction(unittest.TestCase):

    def setUp(self):
        self.config = RattailConfig()

    def test_attributes(self):
        action = monitoring.CommandAction(self.config, "echo test")
        self.assertIs(action.config, self.config)
        self.assertTrue(hasattr(action, 'app'))
        self.assertEqual(action.cmd, "echo test")

    @patch('rattail.monitoring.subprocess')
    def test_run_invokes_command(self, subprocess):
        subprocess.check_call = Mock()
        action = monitoring.CommandAction(self.config, "echo {filename}")
        action('test.txt')
        self.assertEqual(subprocess.check_call.call_count, 1)
        # nb. shell=False is a default kwarg
        subprocess.check_call.assert_called_with(['echo', 'test.txt'], shell=False)

    @patch('rattail.monitoring.subprocess')
    def test_run_with_shell(self, subprocess):
        subprocess.check_call = Mock()
        action = monitoring.CommandAction(self.config, "echo {filename}")
        action('test.txt', shell=True)
        self.assertEqual(subprocess.check_call.call_count, 1)
        subprocess.check_call.assert_called_with('echo test.txt', shell=True)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Generator for 'fabric' projects
"""

import os

import colander

from rattail.projects import PythonProjectGenerator


class FabricProjectGenerator(PythonProjectGenerator):
    """
    Generator for projects meant only to manage Fabric machine
    deployment logic, i.e. no Rattail app.
    """
    key = 'fabric'

    def make_schema(self, **kwargs):
        schema = super(FabricProjectGenerator, self).make_schema(**kwargs)

        schema.add(colander.SchemaNode(name='organization',
                                       typ=colander.String()))

        # TODO: add validation for this?
        schema.add(colander.SchemaNode(name='integrates_with',
                                       typ=colander.String(),
                                       missing=''))

        return schema

    def normalize_context(self, context):
        context = super(FabricProjectGenerator, self).normalize_context(context)

        if not context.get('description'):
            context['description'] = "Fabric project for {}".format(
                context['organization'])

        context['requires']['rattail-fabric2'] = True

        if context['integrates_with'] == 'catapult':
            context['requires']['tailbone-onager'] = True

        return context

    def generate_project(self, output, context, **kwargs):
        super(FabricProjectGenerator, self).generate_project(
            output, context, **kwargs)

        package = os.path.join(output, context['pkg_name'])

        ##############################
        # machines
        ##############################

        machines = os.path.join(output, 'machines')
        os.makedirs(machines)

        ##############################
        # generic-server
        ##############################

        generic_server = os.path.join(machines, 'generic-server')
        os.makedirs(generic_server)

        self.generate('machines/generic-server/README.md.mako',
                      os.path.join(generic_server, 'README.md'),
                      context)

        self.generate('machines/generic-server/Vagrantfile.mako',
                      os.path.join(generic_server, 'Vagrantfile'),
                      context)

        self.generate('machines/generic-server/fabenv.py.dist.mako',
                      os.path.join(generic_server, 'fabenv.py.dist'),
                      context)

        self.generate('machines/generic-server/fabric.yaml.dist',
                      os.path.join(generic_server, 'fabric.yaml.dist'))

        self.generate('machines/generic-server/fabfile.py.mako',
                      os.path.join(generic_server, 'fabfile.py'),
                      context)

        ##############################
        # theo-server
        ##############################

        theo_server = os.path.join(machines, 'theo-server')
        os.makedirs(theo_server)

        self.generate('machines/theo-server/README.md',
                      os.path.join(theo_server, 'README.md'))

        self.generate('machines/theo-server/Vagrantfile',
                      os.path.join(theo_server, 'Vagrantfile'))

        self.generate('machines/theo-server/fabenv.py.dist.mako',
                      os.path.join(theo_server, 'fabenv.py.dist'),
                      context)

        self.generate('machines/theo-server/fabric.yaml.dist',
                      os.path.join(theo_server, 'fabric.yaml.dist'))

        self.generate('machines/theo-server/fabfile.py.mako',
                      os.path.join(theo_server, 'fabfile.py'),
                      context)

        theo_deploy = os.path.join(theo_server, 'deploy')
        os.makedirs(theo_deploy)

        theo_python = os.path.join(theo_deploy, 'python')
        os.makedirs(theo_python)

        self.generate('machines/theo-server/deploy/python/pip.conf.mako',
                      os.path.join(theo_python, 'pip.conf.mako'),
                      context)

        theo_rattail = os.path.join(theo_deploy, 'rattail')
        os.makedirs(theo_rattail)

        self.generate('machines/theo-server/deploy/rattail/rattail.conf.mako',
                      os.path.join(theo_rattail, 'rattail.conf.mako'),
                      context)

        self.generate('machines/theo-server/deploy/rattail/freetds.conf.mako_',
                      os.path.join(theo_rattail, 'freetds.conf.mako'))

        self.generate('machines/theo-server/deploy/rattail/odbc.ini',
                      os.path.join(theo_rattail, 'odbc.ini'))

        theo_theo_common = os.path.join(theo_deploy, 'theo-common')
        os.makedirs(theo_theo_common)

        self.generate('machines/theo-server/deploy/theo-common/rattail.conf.mako',
                      os.path.join(theo_theo_common, 'rattail.conf.mako'),
                      context)

        self.generate('machines/theo-server/deploy/theo-common/web.conf.mako',
                      os.path.join(theo_theo_common, 'web.conf.mako'),
                      context)

        self.generate('machines/theo-server/deploy/theo-common/upgrade.sh.mako',
                      os.path.join(theo_theo_common, 'upgrade.sh.mako'),
                      context)

        self.generate('machines/theo-server/deploy/theo-common/tasks.py.mako_',
                      os.path.join(theo_theo_common, 'tasks.py.mako'))

        self.generate('machines/theo-server/deploy/theo-common/upgrade-wrapper.sh.mako_',
                      os.path.join(theo_theo_common, 'upgrade-wrapper.sh.mako'))

        self.generate('machines/theo-server/deploy/theo-common/supervisor.conf.mako_',
                      os.path.join(theo_theo_common, 'supervisor.conf.mako'))

        self.generate('machines/theo-server/deploy/theo-common/sudoers.mako_',
                      os.path.join(theo_theo_common, 'sudoers.mako'))



# -*- coding: utf-8; -*-
"""add user.local_only

Revision ID: 4f026ae58751
Revises: 64d163d9de3a
Create Date: 2019-10-04 19:00:05.519055

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '4f026ae58751'
down_revision = '64d163d9de3a'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # person
    op.add_column('person', sa.Column('local_only', sa.Boolean(), nullable=True))
    op.add_column('person_version', sa.Column('local_only', sa.Boolean(), autoincrement=False, nullable=True))

    # user
    op.add_column('user', sa.Column('local_only', sa.Boolean(), nullable=True))
    op.add_column('user_version', sa.Column('local_only', sa.Boolean(), autoincrement=False, nullable=True))


def downgrade():

    # user
    op.drop_column('user_version', 'local_only')
    op.drop_column('user', 'local_only')

    # person
    op.drop_column('person_version', 'local_only')
    op.drop_column('person', 'local_only')



# -*- coding: utf-8; -*-
"""add Product.sale_price, tpr_price

Revision ID: 7d0c93d3ff1e
Revises: c45017b36f25
Create Date: 2020-01-03 17:59:04.746965

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '7d0c93d3ff1e'
down_revision = 'c45017b36f25'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # product
    op.add_column('product', sa.Column('sale_price_uuid', sa.String(length=32), nullable=True))
    op.add_column('product', sa.Column('tpr_price_uuid', sa.String(length=32), nullable=True))
    op.create_foreign_key('product_fk_sale_price', 'product', 'product_price', ['sale_price_uuid'], ['uuid'], use_alter=True)
    op.create_foreign_key('product_fk_tpr_price', 'product', 'product_price', ['tpr_price_uuid'], ['uuid'], use_alter=True)
    op.add_column('product_version', sa.Column('sale_price_uuid', sa.String(length=32), autoincrement=False, nullable=True))
    op.add_column('product_version', sa.Column('tpr_price_uuid', sa.String(length=32), autoincrement=False, nullable=True))


def downgrade():

    # product
    op.drop_column('product_version', 'tpr_price_uuid')
    op.drop_column('product_version', 'sale_price_uuid')
    op.drop_constraint('product_fk_tpr_price', 'product', type_='foreignkey')
    op.drop_constraint('product_fk_sale_price', 'product', type_='foreignkey')
    op.drop_column('product', 'tpr_price_uuid')
    op.drop_column('product', 'sale_price_uuid')



# -*- coding: utf-8 -*-
"""add inventory unit cost

Revision ID: aa5962fb998e
Revises: 2fe69ea73233
Create Date: 2017-07-11 22:14:31.878440

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'aa5962fb998e'
down_revision = u'2fe69ea73233'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_inventory_row
    op.add_column('batch_inventory_row', sa.Column('unit_cost', sa.Numeric(precision=9, scale=5), nullable=True))


def downgrade():

    # batch_inventory_row
    op.drop_column('batch_inventory_row', 'unit_cost')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Luigi Tasks
"""

from __future__ import unicode_literals, absolute_import

import os
import datetime
import subprocess
import sys

import luigi
import six

from rattail.time import date_range

# TODO: deprecate / remove this
from rattail.luigi.logging import WarnSummaryIfProblems


class OvernightTask(luigi.Task):
    """
    Base class for overnight automation tasks.
    """
    date = luigi.DateParameter()

    # TODO: subclass must define this
    filename = None

    # how long should we wait after task completes, for datasync to catch up?
    datasync_wait_minutes = None

    def output(self):
        return luigi.LocalTarget('{}/{}'.format(self.date.strftime('%Y/%m/%d'), self.filename))

    def run_command(self):
        raise NotImplementedError

    def touch_output(self):
        with self.output().open('w') as f:
            pass

    def datasync_wait(self, minutes=None, initial_delay=10):
        """
        :param int initial_delay: Number of seconds for the initial delay,
           before we actually start to wait on the queue to clear out.
        """
        if minutes is None:
            minutes = self.datasync_wait_minutes or 10

        # sometimes the datasync queue can take a moment to actually "fill up"
        # initially, so we wait a full minute to ensure that happens before we
        # start actually waiting for it to empty out again
        if initial_delay:
            subprocess.check_call(['sleep', six.text_type(initial_delay)])

        subprocess.check_call([
            'bin/rattail',
            '--config=app/cron.conf',
            'datasync',
            '--timeout={}'.format(minutes),
            'wait',
        ])

    def run(self):
        workdir = os.getcwd()
        os.chdir(sys.prefix)
        self.date_plus = self.date + datetime.timedelta(days=1)
        self.run_command()
        if self.datasync_wait_minutes:
            self.datasync_wait(initial_delay=60)
        os.chdir(workdir)
        self.touch_output()


class OvernightTaskWrapper(luigi.WrapperTask):
    """
    Wrapper task for overnight automation
    """
    date = luigi.DateParameter()


class BackfillTask(luigi.Task):
    """
    Base class for backfill "single date" automation tasks.
    """
    key = luigi.Parameter()
    date = luigi.DateParameter()

    def output(self):
        return luigi.LocalTarget('backfill/{}/{}'.format(
            self.key,
            self.date.strftime('%Y-%m-%d')))

    @property
    def rattail_task(self):
        if not hasattr(self, '_rattail_task'):
            app = self.config.get_app()
            luigi_handler = app.get_luigi_handler()
            self._rattail_task = luigi_handler.get_backfill_task(self.key)
        return self._rattail_task

    def get_script(self):
        if not hasattr(self, 'script'):
            self.script = self.rattail_task['script']
        return self.script

    def run(self):
        cmd = "{} {}".format(self.get_script(), self.date)
        subprocess.check_call(cmd, shell=True)
        self.touch_output()

        # record last date for which this backfill task was ran
        app = self.config.get_app()
        luigi_handler = app.get_luigi_handler()
        luigi_handler.record_backfill_last_date(self.rattail_task, self.date)

    def touch_output(self):
        with self.output().open('w') as f:
            pass


class BackwardBackfillRange(luigi.WrapperTask):
    key = luigi.Parameter()
    start_date = luigi.DateParameter()
    end_date = luigi.DateParameter()

    def requires(self):

        # make sure start date comes before end date
        if self.start_date > self.end_date:
            start_date = self.end_date
            end_date = self.start_date
        else:
            start_date = self.start_date
            end_date = self.end_date

        # get full list of dates for range (inclusive!)
        dates = list(date_range(start_date, end_date))
        # nb. also reverse the list
        dates.sort(reverse=True)

        # require task for each date in range
        # nb. must declare oldest date to prevent infinite loop
        BackwardBackfillTask.newest_date = end_date
        return [BackwardBackfillTask(self.key, date)
                for date in dates]


class BackwardBackfillTask(BackfillTask):

    # nb. caller must explicitly set `newest_date` attribute on the
    # class before using - otherwise an infinite loop is likely

    def requires(self):
        if self.date < self.newest_date:
            return BackwardBackfillTask(self.key,
                                        self.date + datetime.timedelta(days=1))


class ForwardBackfillRange(luigi.WrapperTask):
    key = luigi.Parameter()
    start_date = luigi.DateParameter()
    end_date = luigi.DateParameter()

    def requires(self):

        # make sure start date comes before end date
        if self.start_date > self.end_date:
            start_date = self.end_date
            end_date = self.start_date
        else:
            start_date = self.start_date
            end_date = self.end_date

        # get full list of dates for range (inclusive!)
        dates = list(date_range(start_date, end_date))

        # require task for each date in range
        # nb. must declare oldest date to prevent infinite loop
        ForwardBackfillTask.oldest_date = start_date
        return [ForwardBackfillTask(self.key, date)
                for date in dates]


class ForwardBackfillTask(BackfillTask):

    # nb. caller must explicitly set `oldest_date` attribute on the
    # class before using - otherwise an infinite loop is likely

    def requires(self):
        if self.date > self.oldest_date:
            return ForwardBackfillTask(self.key,
                                       self.date - datetime.timedelta(days=1))



# -*- coding: utf-8; -*-
"""add upgrade.system

Revision ID: 363a96293285
Revises: 0848a1761233
Create Date: 2022-08-19 18:00:43.843663

"""

# revision identifiers, used by Alembic.
revision = '363a96293285'
down_revision = '0848a1761233'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # upgrade
    op.add_column('upgrade', sa.Column('system', sa.String(length=100), nullable=True))


def downgrade():

    # upgrade
    op.drop_column('upgrade', 'system')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handler for Generating Projects
"""

import os
import shutil
import warnings
import zipfile
from collections import OrderedDict

from rattail.util import load_entry_points


class ProjectHandler(object):
    """
    Base class for project handlers.
    """

    def __init__(self, config):
        self.config = config
        self.app = self.config.get_app()

    def get_all_project_generators(self):
        """
        Returns an ``OrderedDict`` with all available project
        generators.
        """
        generators = load_entry_points('rattail.projects')
        generators = sorted(generators.items(),
                            key=lambda itm: itm[0])
        return OrderedDict(generators)

    def get_all_project_types(self):
        """
        Returns the list of keys for *all* possible project types.
        """
        warnings.warn("get_all_project_types() is deprecated; "
                      "please use get_all_project_generators() instead",
                      DeprecationWarning, stacklevel=2)

        return list(self.get_all_project_generators())

    def get_supported_project_generators(self):
        """
        Returns the list of "supported" project generators.
        """
        return self.get_all_project_generators()

    def get_supported_project_types(self):
        """
        Returns the list of keys for "supported" project types.
        """
        warnings.warn("get_supported_project_types() is deprecated; "
                      "please use get_supported_project_generators() instead",
                      DeprecationWarning, stacklevel=2)

        return list(self.get_supported_project_generators())

    def get_project_generator(self, key, require=False):
        """
        Returns a ``ProjectGenerator`` instance for the given key.

        If the key is not valid, returns ``None`` unless
        ``require=True`` in which case an error is raised.
        """
        generators = self.get_all_project_generators()
        if key in generators:
            return generators[key](self.config)
        if require:
            raise RuntimeError("Project generator not found for: {}".format(key))

    def get_storage_dir(self):
        """
        Returns the path to root storage (output) dir for all generated
        projects.
        """
        path = self.config.get('rattail', 'generated_projects.storage_dir')
        if path:
            return path
        return os.path.join(self.config.workdir(require=True),
                            'generated-projects')

    def make_project_schema(self, key):
        """
        Make and return a colander schema representing the context
        needed for generating a project for the given key.
        """
        generator = self.get_project_generator(key, require=True)
        return generator.make_schema()

    def generate_project(self, key, output=None, context=None, **kwargs):
        """
        Generate source code for a new project, and return the path to
        the output folder.

        :param key: Key identifying which type of project to generate.

        :param output: Optional path to the output folder.  If not
           specified, one will be determined automatically.

        :param context: Data dictionary with template context,
           appropriate for the project type.
        """
        generator = self.get_project_generator(key)
        context = generator.normalize_context(context or {})

        if not output:
            folder = context.get('folder', key)
            output = os.path.join(self.get_storage_dir(), folder)

        if os.path.exists(output):
            shutil.rmtree(output)
        os.makedirs(output)

        generator.generate_project(output, context)
        return output

    def zip_output(self, output, zipped=None):
        """
        Compress the given output folder and save as ZIP file.

        :param output: Path to the output folder.

        :param zipped: Optional path to the final ZIP file.  If not
           specified, it will be the same path as ``output`` but with
           a ``.zip`` file extension.
        """
        if not zipped:
            zipped = '{}.zip'.format(output)

        folder = os.path.basename(output)

        with zipfile.ZipFile(zipped, 'w', zipfile.ZIP_DEFLATED) as z:
            self.zipdir(z, output, folder)

        return zipped

    def zipdir(self, zipf, path, folder):
        for root, dirs, files in os.walk(path):
            relative_root = os.path.join(folder, root[len(path)+1:])
            for fname in files:
                zipf.write(os.path.join(root, fname),
                           arcname=os.path.join(relative_root, fname))


class RattailProjectHandler(ProjectHandler):

    def __init__(self, *args, **kwargs):
        warnings.warn("RattailProjectHandler is deprecated; "
                      "please just use ProjectHandler instead",
                      DeprecationWarning, stacklevel=2)
        super(RattailProjectHandler, self).__init__(*args, **kwargs)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Clientele Handler
"""

from collections import OrderedDict
import warnings

from rattail.util import load_object
from rattail.app import GenericHandler


class ClienteleHandler(GenericHandler):
    """
    Base class and default implementation for clientele handlers.
    """

    def choice_uses_dropdown(self):
        """
        Returns boolean indicating whether a customer choice should be
        presented to the user via a dropdown (select) element, vs.  an
        autocomplete field.  The latter is the default because
        potentially the customer list can be quite large, so we avoid
        loading them all in the dropdown unless so configured.

        :returns: Boolean; if true then a dropdown should be used;
           otherwise (false) autocomplete is used.
        """
        return self.config.getbool('rattail', 'customers.choice_uses_dropdown',
                                   default=False)

    def ensure_customer(self, person):
        """
        Returns the customer record associated with the given person, creating
        it first if necessary.
        """
        customer = self.get_customer(person)
        if customer:
            return customer

        session = self.get_session(person)
        customer = self.make_customer(person)
        session.add(customer)
        session.flush()
        session.refresh(person)
        return customer

    def get_customer(self, obj):
        """
        Return the Customer associated with the given object, if any.
        """
        model = self.model

        if isinstance(obj, model.Customer):
            return obj

        else:
            person = self.app.get_person(obj)
            if person:
                # TODO: all 3 options below are indeterminate, since it's
                # *possible* for a person to hold multiple accounts
                # etc. but not sure how to fix in a generic way?  maybe
                # just everyone must override as needed
                if person.customer_accounts:
                    return person.customer_accounts[0]
                for shopper in person.customer_shoppers:
                    if shopper.shopper_number == 1:
                        return shopper.customer
                # legacy fallback
                if person.customers:
                    return person.customers[0]

    def get_email_address(self, customer, **kwargs):
        """
        Return the first email address found for the given customer.

        :returns: The email address as string, or ``None``.
        """
        warnings.warn("clientele.get_email_address(customer) is deprecated; please "
                      "use app.get_contact_email_address(customer) instead",
                      DeprecationWarning, stacklevel=2)
        return self.app.get_contact_email_address(customer)

    def get_customers_for_account_holder(
            self,
            person,
            **kwargs
    ):
        """
        Return all Customer records for which the given Person is the
        account holder.
        """
        customers = OrderedDict()

        # find customers for which person is account holder
        for customer in person.customer_accounts:
            customers.setdefault(customer.uuid, customer)

        # find customers for which person is primary shopper
        for shopper in person.customer_shoppers:
            if shopper.shopper_number == 1:
                customer = shopper.customer
                customers.setdefault(customer.uuid, customer)

        # nb. legacy
        for customer in person.customers:
            customers.setdefault(customer.uuid, customer)

        return list(customers.values())

    def get_active_shopper(
            self,
            customer,
            **kwargs
    ):
        """
        Return the "active" shopper record for the given customer.

        This should never return multiple shoppers, either one or none.
        """
        for shopper in customer.shoppers:
            if shopper.active:
                return shopper

    def get_person(self, customer):
        """
        Returns the person associated with the given customer, if there is one.
        """
        warnings.warn("ClienteleHandler.get_person() is deprecated; "
                      "please use AppHandler.get_person() instead")

        return self.app.get_person(customer)

    def should_use_legacy_people(self):
        avoid = self.config.getbool('rattail',
                                    'customer_accounts.avoid_legacy_people',
                                    default=False)
        return not avoid

    def make_customer(self, person, **kwargs):
        """
        Create and return a new customer record.
        """
        customer = self.model.Customer()
        customer.name = person.display_name
        if self.should_use_legacy_people():
            customer.people.append(person)
        else:
            customer.account_holder = person
        return customer

    def get_first_phone(self, customer, **kwargs):
        """
        Return the first available phone record found, either for the
        customer, or its first person.
        """
        phone = customer.first_phone()
        if phone:
            return phone

        person = self.app.get_person(customer)
        if person:
            return person.first_phone()

    def get_first_phone_number(self, customer, **kwargs):
        """
        Return the first available phone number found, either for the
        customer, or its first person.
        """
        phone = self.get_first_phone(customer)
        if phone:
            return phone.number

    def get_first_email(self, customer, invalid=False, **kwargs):
        """
        Return the first available email record found, either for the
        customer, or its first person.
        """
        email = customer.first_email(invalid=invalid)
        if email:
            return email

        person = self.app.get_person(customer)
        if person:
            return person.first_email(invalid=invalid)

    def get_first_email_address(self, customer, invalid=False, **kwargs):
        """
        Return the first available email address found, either for the
        customer, or its first person.
        """
        email = self.get_first_email(customer, invalid=invalid)
        if email:
            return email.address


def get_clientele_handler(config, **kwargs):
    """
    Create and return the configured :class:`ClienteleHandler` instance.
    """
    spec = config.get('rattail', 'clientele.handler')
    if spec:
        factory = load_object(spec)
    else:
        factory = ClienteleHandler
    return factory(config, **kwargs)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
FileMon logic for Bouncer
"""

from __future__ import unicode_literals, absolute_import

from rattail.filemon import Action


class HandleBounceFile(Action):
    """
    Fully "handle" the given email message file, acting on it as
    needed, depending on whether or not it is actually considered
    to be a bounce etc.
    """
    
    def __call__(self, path, key='default', **kwargs):
        handler = self.app.get_bounce_handler(key)
        handler.handle_bounce_file(path)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Label Printing
"""

import io
import os
import os.path
import socket
import shutil
from collections import OrderedDict

from rattail.app import GenericHandler
from rattail.core import Object
from rattail.files import temp_path
from rattail.exceptions import LabelPrintingError
from rattail.time import localtime


class LabelHandler(GenericHandler):
    """
    Base class and default implementation for label handlers.
    """

    def get_formatter(self, profile, ignore_errors=False):
        """
        Return the label formatter for the given profile.

        :param profile: A
           :class:`~rattail.db.model.labels.LabelProfile` instance.

        :returns: A :class:`~rattail.labels.LabelFormatter` instance.
        """
        if not profile.formatter_spec:
            if ignore_errors:
                return
            raise ValueError("label profile has no formatter_spec: {}".format(
                profile))

        factory = self.app.load_object(profile.formatter_spec)
        formatter = factory(self.config, template=profile.format)
        return formatter

    def get_printer(self, profile, ignore_errors=False):
        """
        Return the label printer for the given profile.

        :param profile: A
           :class:`~rattail.db.model.labels.LabelProfile` instance.

        :returns: A :class:`~rattail.labels.LabelPrinter` instance.
        """
        if not profile.printer_spec:
            if ignore_errors:
                return
            raise ValueError("label profile has no printer_spec: {}".format(
                profile))

        # create the printer
        factory = self.app.load_object(profile.printer_spec)
        printer = factory(self.config)

        # establish settings for it
        for name in printer.required_settings:
            value = self.get_printer_setting(profile, name)
            setattr(printer, name, value)

        # give it a formatter
        printer.formatter = self.get_formatter(profile,
                                               ignore_errors=ignore_errors)

        return printer

    def get_printer_setting(self, profile, name):
        """
        Read a printer setting from the DB.
        """
        if not profile.uuid:
            return
        session = self.app.get_session(profile)
        name = 'labels.{}.printer.{}'.format(profile.uuid, name)
        return self.app.get_setting(session, name)

    def save_printer_setting(self, profile, name, value):
        """
        Write a printer setting to the DB.
        """
        session = self.app.get_session(profile)
        if not profile.uuid:
            session.flush()
        name = 'labels.{}.printer.{}'.format(profile.uuid, name)
        self.app.save_setting(session, name, value)

    def get_product_label_data(self, product, **kwargs):
        """
        Return a data dict with common product fields, suitable for
        use with formatting labels for printing.

        The intention is that this method should be able to provide
        enough data to make basic label printing possible.  In fact
        when printing is done "ad-hoc" for one product at a time, the
        data used for printing comes only from this method.

        :param product: A :class:`~rattail.db.model.products.Product`
           instance from which field values will come.

        :returns: Data dict containing common product fields.
        """
        data = {
            'description': product.description,
            'size': product.size,
        }

        brand = product.brand
        data['brand_name'] = brand.name if brand else None

        regprice = product.regular_price
        data['regular_price'] = regprice.price if regprice else None
        data['regular_pack_price'] = regprice.pack_price if regprice else None

        tprprice = product.tpr_price
        data['tpr_price'] = tprprice.price if tprprice else None
        data['tpr_starts'] = None
        if tprprice and tprprice.starts:
            starts = self.app.localtime(tprprice.starts, from_utc=True)
            data['tpr_starts'] = self.app.render_datetime(tprprice.starts)
        data['tpr_ends'] = None
        if tprprice and tprprice.ends:
            ends = self.app.localtime(tprprice.ends, from_utc=True)
            data['tpr_ends'] = self.app.render_datetime(tprprice.ends)

        salprice = product.sale_price
        data['sale_price'] = salprice.price if salprice else None
        data['sale_starts'] = None
        if salprice and salprice.starts:
            starts = self.app.localtime(salprice.starts, from_utc=True)
            data['sale_starts'] = self.app.render_datetime(salprice.starts)
        data['sale_ends'] = None
        if salprice and salprice.ends:
            ends = self.app.localtime(salprice.ends, from_utc=True)
            data['sale_ends'] = self.app.render_datetime(salprice.ends)

        curprice = product.current_price
        data['current_price'] = curprice.price if curprice else None
        data['current_starts'] = None
        if curprice and curprice.starts:
            starts = self.app.localtime(curprice.starts, from_utc=True)
            data['current_starts'] = self.app.render_datetime(curprice.starts)
        data['current_ends'] = None
        if curprice and curprice.ends:
            ends = self.app.localtime(curprice.ends, from_utc=True)
            data['current_ends'] = self.app.render_datetime(curprice.ends)

        sugprice = product.suggested_price
        data['suggested_price'] = sugprice.price if sugprice else None

        vendor = product.vendor
        data['vendor_name'] = vendor.name if vendor else None

        return data


class LabelPrinter(object):
    """
    Base class for all label printers.

    Label printing devices which are "natively" supported by Rattail will each
    derive from this class in order to provide implementation details specific
    to the device.  You will typically instantiate one of those subclasses (or
    one of your own design) in order to send labels to your physical printer.
    """

    profile_name = None
    formatter = None
    required_settings = None

    def __init__(self, config):
        self.config = config

    def print_labels(self, labels, *args, **kwargs):
        """
        Prints labels found in ``labels``.
        """
        raise NotImplementedError


class CommandPrinter(LabelPrinter):
    """
    Generic :class:`LabelPrinter` class which "prints" labels via native
    printer (textual) commands.  It does not directly implement any method for
    sending the commands to a printer; a subclass must be used for that.
    """

    def batch_header_commands(self):
        """
        This method, if implemented, must return a sequence of string commands
        to be interpreted by the printer.  These commands will be the first
        which are written to the file.
        """

        return None

    def batch_footer_commands(self):
        """
        This method, if implemented, must return a sequence of string commands
        to be interpreted by the printer.  These commands will be the last
        which are written to the file.
        """

        return None


class CommandFilePrinter(CommandPrinter):
    """
    Generic :class:`LabelPrinter` implementation which "prints" labels to a
    file in the form of native printer (textual) commands.  The output file is
    then expected to be picked up by a file monitor, and finally sent to the
    printer from there.
    """

    required_settings = {'output_dir': "Output Folder"}
    output_dir = None

    def print_labels(self, labels, output_dir=None, progress=None):
        """
        "Prints" ``labels`` by generating a command file in the output folder.
        The full path of the output file to which commands are written will be
        returned to the caller.

        If ``output_dir`` is not specified, and the printer instance is
        associated with a :class:`LabelProfile` instance, then config will be
        consulted for the output path.  If a path still is not found, the
        current (working) directory will be assumed.
        """

        if not output_dir:
            output_dir = self.output_dir
        if not output_dir:
            raise LabelPrintingError("Printer does not have an output folder defined")

        labels_path = temp_path(prefix='rattail.', suffix='.labels')
        labels_file = open(labels_path, 'w')

        header = self.batch_header_commands()
        if header:
            labels_file.write('%s\n' % '\n'.join(header))

        commands = self.formatter.format_labels(labels, progress=progress)
        if commands is None:
            labels_file.close()
            os.remove(labels_path)
            return None

        labels_file.write(commands)

        footer = self.batch_footer_commands()
        if footer:
            labels_file.write('%s\n' % '\n'.join(footer))

        labels_file.close()
        fn = '{0}_{1}.labels'.format(
            socket.gethostname(),
            localtime(self.config).strftime('%Y-%m-%d_%H-%M-%S'))
        final_path = os.path.join(output_dir, fn)
        shutil.move(labels_path, final_path)
        return final_path


# Force ordering for network printer required settings.
settings = OrderedDict()
settings['address'] = "IP Address"
settings['port'] = "Port"
settings['timeout'] = "Timeout"

class CommandNetworkPrinter(CommandPrinter):
    """
    Generic :class:`LabelPrinter` implementation which "prints" labels to a
    network socket in the form of native printer (textual) commands.  The
    socket is assumed to exist on a networked label printer.
    """

    required_settings = settings
    address = None
    port = None
    timeout = None

    def print_labels(self, labels, progress=None):
        """
        Prints ``labels`` by generating commands and sending directly to a
        socket which exists on a networked printer.
        """

        if not self.address:
            raise LabelPrintingError("Printer does not have an IP address defined")
        if not self.port:
            raise LabelPrintingError("Printer does not have a port defined.")

        data = io.StringIO()

        header = self.batch_header_commands()
        if header:
            header = "{0}\n".format('\n'.join(header))
            data.write(header.encode('utf_8'))

        commands = self.formatter.format_labels(labels, progress=progress)
        if commands is None: # process canceled?
            data.close()
            return None
        data.write(commands.encode('utf_8'))

        footer = self.batch_footer_commands()
        if footer:
            footer = "{0}\n".format('\n'.join(footer))
            data.write(footer.encode('utf_8'))

        try:
            timeout = int(self.timeout)
        except ValueError:
            timeout = socket.getdefaulttimeout()

        try:
            # Must pass byte-strings (not unicode) to this function.
            sock = socket.create_connection((self.address.decode('utf_8'), int(self.port)), timeout)
            bytes = sock.send(data.getvalue())
            sock.close()
            return bytes
        finally:
            data.close()


class LabelFormatter(Object):
    """
    Base class for all label formatters.
    """
    template = None

    def __init__(self, config, template=None):
        self.config = config
        self.app = self.config.get_app()
        if template:
            self.template = template

    @property
    def default_template(self):
        """
        Default formatting template.  This will be used if no template
        is defined within the label profile; see also
        :attr:`rattail.db.model.labels.LabelProfile.format`.
        """
        raise NotImplementedError

    def format_labels(self, labels, progress=None, **kwargs):
        """
        Formats a set of labels and returns the result.

        :param labels: Sequence of 2-tuples representing labels to be
           formatted, and ultimately printed.
        """
        raise NotImplementedError


class CommandFormatter(LabelFormatter):
    """
    Base class and default implementation for label formatters which
    generate raw printer commands for sending (in)directly to the
    printer device.

    There are various printer command languages out there; this class
    is not language-specific and should work for any of them.

    .. attribute:: template

       Formatting template.  This is a string containing a template of
       raw printer commands, suitable for printing a single label
       record.  Value for this normally comes from
       :attr:`rattail.db.model.labels.LabelProfile.format`.

       Normally these commands would print a "complete" label in terms
       of physical media, but not so for 2-up, in which case the
       template should only contain commands for "half" the label,
       i.e. only the commands to print one "record".

       There are 2 "languages" at play within the template:

       * template language
       * printer command language

       The template language refers to the syntax of the template
       itself, which ultimately will be "rendered" into a final result
       which should contain valid printer command language.  (See also
       :class:`~rattail.labels.CommandFormatter.format_labels()`.)
       Thus far there is only one template language supported,
       although it is likely more will be added in the future:

       * :ref:`python:old-string-formatting`

       The printer command language refers to the syntax of commands
       which can be sent to the printer in order to cause it to
       produce desired physical media, i.e. "formatted printed label".
       There are a number of printer command languages out there; the
       one you need to use will depend on the make and/or model and/or
       settings for your printer device.  Thus far the following
       languages have been used successfully:

       * `Cognitive Programming Language (CPL) <http://cognitivetpg.com/assets/downloads/105-008-04%20F(CPL%20ProgrammersGuide).pdf>`_
       * `Zebra Programming Language (ZPL) <https://en.wikipedia.org/wiki/Zebra_Programming_Language>`_

       A template example using ZPL:

       .. code-block:: none

          ^XA
          ^FO035,65^A0N,40,30^FD%(brand)-17.17s^FS
          ^FO035,110^A0N,30,30^FD%(description)s %(size)s^FS
          ^FO163,36^A0N,80,55^FB230,1,0,R,0^FD$%(price)0.2f^FS
          ^FO265,170,0^A0N,25,20^FD%(vendor)-14.14s^FS
          ^FO050,144^BY2%(barcode)s^FS
          ^XZ

       A template example using CPL for a 2-up layout:

       .. code-block:: none

          STRING 5X7 %(description_x)d 5 %(description1)s
          STRING 5X7 %(description_x)d 15 %(description2)s
          BARCODE %(barcode_type)s %(barcode_x)d 60 20 %(barcode)s
    """

    def format_labels(self, labels, progress=None, **kwargs):
        """
        Format a set of labels and return the result.  By "formatting"
        here we really mean generating a set of commands which
        ultimately will be submitted directly to a label printer
        device.

        Each of the ``labels`` specified should be a 2-tuple like
        ``(data, quantity)``, where ``data`` is a dict of record data
        (e.g. product description, price etc.) and ``quantity`` is the
        number of labels to be printed for that record.

        The formatter's :attr:`template` is "rendered" by feeding it
        the data dict from a single label record.  That process is
        repeated until all labels have been rendered.

        Note that the formatting template is only able to reference
        fields present in the ``data`` dict for any given label
        record.  If the incoming data is incomplete then you can add
        to it by overriding :meth:`get_all_data()`.

        :param labels: Sequence of 2-tuples representing labels to be
           formatted for printing.

        :param progress: Optional progress factory.

        :returns: Unicode string containing the formatted label data,
           i.e. commands to print the labels.
        """
        fmt = io.StringIO()

        def format_label(record, i):
            data, quantity = record
            product = data.get('product')
            if not product:
                return
            for j in range(quantity):
                header = self.label_header_commands()
                if header:
                    header = "{0}\n".format('\n'.join(header))
                    fmt.write(header.encode('utf_8'))
                data = self.get_all_data(data)
                body = "{}\n".format('\n'.join(self.label_body_commands(product, data)))
                fmt.write(body)
                footer = self.label_footer_commands()
                if footer:
                    footer = "{0}\n".format('\n'.join(footer))
                    fmt.write(footer.encode('utf_8'))

        self.app.progress_loop(format_label, labels, progress,
                               message="Formatting labels")

        val = fmt.getvalue()
        fmt.close()
        return val

    def get_all_data(self, data, **kwargs):
        """
        Returns the "complete' data dict for a given label record.

        When the caller asks us to format labels, it provides a data
        dict for each label to be printed.  This method is able to add
        more things to that data dict, if needed.

        Note that which fields are actually needed will depend on the
        contents of :attr:`template`.

        By default this will check the data dict for a ``'product'``
        key, and if there is a value, calls :meth:`get_product_data()`
        to add common product fields.

        :param data: Dict of data for a label record, as provided by
           the caller.

        :returns: Final data dict with all necessary fields.
        """
        if data.get('product'):
            data = self.get_product_data(data, data['product'])
        return data

    def get_product_data(self, data, product, **kwargs):
        """
        Add common product fields to the given data dict.

        The intention is that even if ``data`` is an empty dict, this
        method should be able to add enough data to make basic label
        printing possible.

        Default logic for this will call
        :meth:`rattail.labels.LabelHandler.get_product_label_data()`
        to get the product data dict, and then use ``data`` from the
        caller to override anything as needed, and return the result.

        :param data: Dict of data for a label record.

        :param product: A :class:`~rattail.db.model.products.Product`
           instance to which the label record applies, and from which
           additional field values will come.

        :returns: Final data dict including common product fields.
        """
        label_handler = self.app.get_label_handler()
        final_data = label_handler.get_product_label_data(product)
        final_data.update(data)
        return final_data

    def label_header_commands(self):
        """
        This method, if implemented, must return a sequence of string commands
        to be interpreted by the printer.  These commands will immediately
        precede each *label* in one-up printing, and immediately precede each
        *label pair* in two-up printing.
        """

    def label_body_commands(self, product, data, **kwargs):
        raise NotImplementedError

    def label_footer_commands(self):
        """
        This method, if implemented, must return a sequence of string commands
        to be interpreted by the printer.  These commands will immedately
        follow each *label* in one-up printing, and immediately follow each
        *label pair* in two-up printing.
        """


class TwoUpCommandFormatter(CommandFormatter):
    """
    Generic subclass of :class:`LabelFormatter` which generates native printer
    (textual) commands.

    This class contains logic to implement "two-up" label printing.
    """

    @property
    def half_offset(self):
        """
        The X-coordinate value by which the second label should be offset, when
        two labels are printed side-by-side.
        """
        raise NotImplementedError

    def format_labels(self, labels, progress=None, **kwargs):
        ""                      # avoid auto-generated docs
        fmt = io.StringIO()
        self.half_started = False

        def format_label(record, i):
            data, quantity = record
            product = data.get('product')
            if not product:
                return
            for j in range(quantity):
                kw = self.get_all_data(data)
                if self.half_started:
                    kw['x'] = self.half_offset
                    fmt.write('{}\n'.format('\n'.join(self.label_body_commands(product, kw))))
                    footer = self.label_footer_commands()
                    if footer:
                        fmt.write('{}\n'.format('\n'.join(footer)))
                    self.half_started = False
                else:
                    header = self.label_header_commands()
                    if header:
                        fmt.write('{}\n'.format('\n'.join(header)))
                    kw['x'] = 0
                    fmt.write('{}\n'.format('\n'.join(self.label_body_commands(product, kw))))
                    self.half_started = True

        self.app.progress_loop(format_label, labels, progress,
                               message="Formatting labels")

        if self.half_started:
            footer = self.label_footer_commands()
            if footer:
                fmt.write('{}\n'.format('\n'.join(footer)))

        val = fmt.getvalue()
        fmt.close()
        return val



# -*- coding: utf-8 -*-
"""add purchase.po_line_number

Revision ID: a8c96f5491ae
Revises: a47a4d73b662
Create Date: 2016-12-29 15:53:16.596495

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = 'a8c96f5491ae'
down_revision = u'a47a4d73b662'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


batch_id_seq = sa.Sequence('batch_id_seq')


def upgrade():

    # purchase_batch_row
    op.add_column('purchase_batch_row', sa.Column('po_line_number', sa.Integer(), nullable=True))

    # purchase_item
    op.add_column('purchase_item', sa.Column('po_line_number', sa.Integer(), nullable=True))


def downgrade():

    # purchase_item
    op.drop_column('purchase_item', 'po_line_number')

    # purchase_batch_row
    op.drop_column('purchase_batch_row', 'po_line_number')
